import asyncio
import json
import urllib.parse
import warnings
from asyncio import Task
from collections.abc import Coroutine
from datetime import datetime
from typing import Callable, Any, Union, cast
from uuid import UUID

import httpx
import web3
import websockets
from eth_abi.abi import encode
from eth_account.account import Account
from eth_account.datastructures import SignedMessage
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from solders.instruction import Instruction
from solders.pubkey import Pubkey
from solders.sysvar import INSTRUCTIONS
from websockets.client import WebSocketClientProtocol

from express_relay.constants import (
    OPPORTUNITY_ADAPTER_CONFIGS,
    EXECUTION_PARAMS_TYPESTRING,
    SVM_CONFIGS,
)
from express_relay.models.evm import (
    Address,
    Bytes32,
    TokenAmount,
    BidEvm,
)
from express_relay.models.svm import SvmChainUpdate
from express_relay.models import (
    Bid,
    BidStatusUpdate,
    BidResponse,
    OpportunityBidParams,
    OpportunityBid,
    OpportunityParams,
    Opportunity,
    OpportunityRoot,
    ClientMessage,
    BidResponseRoot,
    OpportunityDelete,
    OpportunityDeleteRoot,
)
from express_relay.models.base import UnsupportedOpportunityVersionException
from express_relay.models.evm import OpportunityEvm
from express_relay.svm.generated.express_relay.instructions.submit_bid import submit_bid
from express_relay.svm.generated.express_relay.program_id import (
    PROGRAM_ID as SVM_EXPRESS_RELAY_PROGRAM_ID,
)
from express_relay.svm.generated.express_relay.types.submit_bid_args import (
    SubmitBidArgs,
)
from express_relay.svm.limo_client import LimoClient


def _get_permitted_tokens(
    sell_tokens: list[TokenAmount],
    bid_amount: int,
    call_value: int,
    weth_address: Address,
) -> list[dict[str, Union[str, int]]]:
    """
    Extracts the sell tokens in the permit format.

    Args:
        sell_tokens: A list of TokenAmount objects representing the sell tokens.
        bid_amount: An integer representing the amount of the bid (in wei).
        call_value: An integer representing the call value of the bid (in wei).
        weth_address: The address of the WETH token.
    Returns:
        A list of dictionaries representing the sell tokens in the permit format.
    """
    permitted_tokens: list[dict[str, Union[str, int]]] = [
        {
            "token": token.token,
            "amount": int(token.amount),
        }
        for token in sell_tokens
    ]

    for token in permitted_tokens:
        if token["token"] == weth_address:
            sell_token_amount = cast(int, token["amount"])
            token["amount"] = sell_token_amount + call_value + bid_amount
            return permitted_tokens

    if bid_amount + call_value > 0:
        permitted_tokens.append(
            {
                "token": weth_address,
                "amount": bid_amount + call_value,
            }
        )

    return permitted_tokens


class ExpressRelayClientException(Exception):
    pass


class ExpressRelayClient:
    def __init__(
        self,
        server_url: str,
        api_key: str | None = None,
        opportunity_callback: (
            Callable[[Opportunity], Coroutine[Any, Any, Any]] | None
        ) = None,
        bid_status_callback: (
            Callable[[BidStatusUpdate], Coroutine[Any, Any, Any]] | None
        ) = None,
        svm_chain_update_callback: (
            Callable[[SvmChainUpdate], Coroutine[Any, Any, Any]] | None
        ) = None,
        remove_opportunities_callback: (
            Callable[[OpportunityDelete], Coroutine[Any, Any, Any]] | None
        ) = None,
        timeout_response_secs: int = 10,
        ws_options: dict[str, Any] | None = None,
        http_options: dict[str, Any] | None = None,
    ):
        """
        Args:
            server_url: The URL of the auction server.
            opportunity_callback: An async function that serves as the callback on a new opportunity. Should take in one external argument of type Opportunity.
            bid_status_callback: An async function that serves as the callback on a new bid status update. Should take in one external argument of type BidStatusUpdate.
            svm_chain_update_callback: An async function that serves as the callback on a new svm chain update. Should take in one external argument of type SvmChainUpdate.
            remove_opportunities_callback: An async function that serves as the callback on an opportunities delete. Should take in one external argument of type OpportunityDelete.
            timeout_response_secs: The number of seconds to wait for a response message from the server.
            ws_options: Keyword arguments to pass to the websocket connection.
            http_options: Keyword arguments to pass to the HTTP client.
        """
        parsed_url = urllib.parse.urlparse(server_url)
        if parsed_url.scheme == "https":
            ws_scheme = "wss"
        elif parsed_url.scheme == "http":
            ws_scheme = "ws"
        else:
            raise ValueError("Invalid server URL")

        self.server_url = server_url
        self.api_key = api_key
        self.ws_endpoint = parsed_url._replace(scheme=ws_scheme, path="/v1/ws").geturl()
        self.ws_msg_counter = 0
        self.ws: WebSocketClientProtocol
        self.ws_lock = asyncio.Lock()
        self.ws_loop: Task[Any]
        self.ws_msg_futures: dict[str, asyncio.Future] = {}
        self.timeout_response_secs = timeout_response_secs
        if ws_options is None:
            ws_options = {}
        self.ws_options = ws_options
        if http_options is None:
            http_options = {}
        self.http_options = http_options
        self.opportunity_callback = opportunity_callback
        self.bid_status_callback = bid_status_callback
        self.svm_chain_update_callback = svm_chain_update_callback
        self.remove_opportunities_callback = remove_opportunities_callback
        if self.api_key:
            authorization_header = f"Bearer {self.api_key}"
            if "headers" not in self.http_options:
                self.http_options["headers"] = {}
            self.http_options["headers"]["Authorization"] = authorization_header
            if "extra_headers" not in self.ws_options:
                self.ws_options["extra_headers"] = {}
            self.ws_options["extra_headers"]["Authorization"] = authorization_header

    async def start_ws(self):
        """
        Initializes the websocket connection to the server, if not already connected.
        """
        async with self.ws_lock:
            if not hasattr(self, "ws"):
                self.ws = await websockets.connect(self.ws_endpoint, **self.ws_options)

            if not hasattr(self, "ws_loop"):
                ws_call = self.ws_handler(
                    self.opportunity_callback, self.bid_status_callback,
                    self.svm_chain_update_callback, self.remove_opportunities_callback,
                )
                self.ws_loop = asyncio.create_task(ws_call)

    async def close_ws(self):
        """
        Closes the websocket connection to the server.
        """
        async with self.ws_lock:
            await self.ws.close()

    async def get_ws_loop(self) -> asyncio.Task:
        """
        Returns the websocket handler loop.
        """
        await self.start_ws()

        return self.ws_loop

    def convert_client_msg_to_server(self, client_msg: ClientMessage) -> dict:
        """
        Converts the params of a ClientMessage model dict to the format expected by the server.

        Args:
            client_msg: The message to send to the server.
        Returns:
            The message as a dict with the params converted to the format expected by the server.
        """
        msg = client_msg.model_dump()
        method = msg["params"]["method"]
        msg["id"] = str(self.ws_msg_counter)
        self.ws_msg_counter += 1

        if method == "post_bid":
            msg["params"] = {"bid": msg["params"]}

        msg["method"] = method

        return msg

    async def send_ws_msg(self, client_msg: ClientMessage) -> dict:
        """
        Sends a message to the server via websocket.

        Args:
            client_msg: The message to send.
        Returns:
            The result of the response message from the server.
        """
        await self.start_ws()

        msg = self.convert_client_msg_to_server(client_msg)

        future = asyncio.get_event_loop().create_future()
        self.ws_msg_futures[msg["id"]] = future

        await self.ws.send(json.dumps(msg))

        # await the response for the sent ws message from the server
        msg_response = await asyncio.wait_for(
            future, timeout=self.timeout_response_secs
        )

        return self.process_response_msg(msg_response)

    def process_response_msg(self, msg: dict) -> dict:
        """
        Processes a response message received from the server via websocket.

        Args:
            msg: The message to process.
        Returns:
            The result field of the message.
        """
        if msg.get("status") and msg.get("status") != "success":
            raise ExpressRelayClientException(
                f"Error in websocket response with message id {msg.get('id')}: {msg.get('result')}"
            )
        return msg["result"]

    async def subscribe_chains(self, chain_ids: list[str]):
        """
        Subscribes websocket to a list of chain IDs for new opportunities.

        Args:
            chain_ids: A list of chain IDs to subscribe to.
        """
        params = {
            "method": "subscribe",
            "chain_ids": chain_ids,
        }
        client_msg = ClientMessage.model_validate({"params": params})
        await self.send_ws_msg(client_msg)

    async def unsubscribe_chains(self, chain_ids: list[str]):
        """
        Unsubscribes websocket from a list of chain IDs for new opportunities.

        Args:
            chain_ids: A list of chain IDs to unsubscribe from.
        """
        params = {
            "method": "unsubscribe",
            "chain_ids": chain_ids,
        }
        client_msg = ClientMessage.model_validate({"params": params})
        await self.send_ws_msg(client_msg)

    async def submit_bid(self, bid: Bid, subscribe_to_updates: bool = True) -> UUID:
        """
        Submits a bid to the auction server.

        Args:
            bid: An object representing the bid to submit.
            subscribe_to_updates: A boolean indicating whether to subscribe to the bid status updates.
        Returns:
            The ID of the submitted bid.
        """
        bid_dict = bid.model_dump()
        if subscribe_to_updates:
            bid_dict["method"] = "post_bid"
            client_msg = ClientMessage.model_validate({"params": bid_dict})
            result = await self.send_ws_msg(client_msg)
            bid_id = UUID(result.get("id"))
        else:
            async with httpx.AsyncClient(**self.http_options) as client:
                resp = await client.post(
                    urllib.parse.urlparse(self.server_url)
                    ._replace(path="/v1/bids")
                    .geturl(),
                    json=bid_dict,
                )

            resp.raise_for_status()
            bid_id = UUID(resp.json().get("id"))

        return bid_id

    async def ws_handler(
        self,
        opportunity_callback: (
            Callable[[Opportunity], Coroutine[Any, Any, Any]] | None
        ) = None,
        bid_status_callback: (
            Callable[[BidStatusUpdate], Coroutine[Any, Any, Any]] | None
        ) = None,
        svm_chain_update_callback: (
            Callable[[SvmChainUpdate], Coroutine[Any, Any, Any]] | None
        ) = None,
        remove_opportunities_callback: (
            Callable[[OpportunityDelete], Coroutine[Any, Any, Any]] | None
        ) = None,
    ):
        """
        Continually handles new ws messages as they are received from the server via websocket.

        Args:
            opportunity_callback: An async function that serves as the callback on a new opportunity. Should take in one external argument of type Opportunity.
            bid_status_callback: An async function that serves as the callback on a new bid status update. Should take in one external argument of type BidStatusUpdate.
            svm_chain_update_callback: An async function that serves as the callback on a new svm chain update. Should take in one external argument of type SvmChainUpdate.
            remove_opportunities_callback: An async function that serves as the callback on an opportunities delete. Should take in one external argument of type OpportunityDelete.
        """
        if not self.ws:
            raise ExpressRelayClientException("Websocket not connected")

        async for msg in self.ws:
            msg_json = json.loads(msg)

            if msg_json.get("type"):
                if msg_json.get("type") == "new_opportunity":
                    if opportunity_callback is not None:
                        opportunity = OpportunityRoot.model_validate(
                            msg_json["opportunity"]
                        )
                        if opportunity:
                            asyncio.create_task(opportunity_callback(opportunity.root))

                elif msg_json.get("type") == "bid_status_update":
                    if bid_status_callback is not None:
                        bid_status_update = BidStatusUpdate.model_validate(
                            msg_json["status"]
                        )
                        asyncio.create_task(bid_status_callback(bid_status_update))

                elif msg_json.get("type") == "svm_chain_update":
                    if svm_chain_update_callback is not None:
                        svm_chain_update = SvmChainUpdate.model_validate(
                            msg_json["update"]
                        )
                        asyncio.create_task(svm_chain_update_callback(svm_chain_update))

                elif msg_json.get("type") == "remove_opportunities":
                    if remove_opportunities_callback is not None:
                        remove_opportunities = OpportunityDeleteRoot.model_validate(
                            msg_json["opportunity_delete"]
                        )
                        if remove_opportunities:
                            asyncio.create_task(remove_opportunities_callback(remove_opportunities))

            elif msg_json.get("id"):
                future = self.ws_msg_futures.pop(msg_json["id"])
                future.set_result(msg_json)

    async def get_opportunities(self, chain_id: str | None = None, from_time: datetime | None = None, limit: int | None = None) -> list[Opportunity]:
        """
        Connects to the server and fetches opportunities.

        Args:
            chain_id: The chain ID to fetch opportunities for. If None, fetches opportunities across all chains.
            from_time: The datetime to fetch opportunities from. If None, fetches from the beginning of time.
            limit: Number of opportunities to fetch. If None, uses the default server limit.
        Returns:
            A list of opportunities.
        """
        params = {}
        if chain_id:
            params["chain_id"] = chain_id
        if from_time:
            params["from_time"] = from_time.astimezone().isoformat(timespec="microseconds")
        if limit:
            params["limit"] = str(limit)

        async with httpx.AsyncClient(**self.http_options) as client:
            resp = await client.get(
                urllib.parse.urlparse(self.server_url)
                ._replace(path="/v1/opportunities")
                .geturl(),
                params=params,
            )

        resp.raise_for_status()

        opportunities: list[Opportunity] = []
        for opportunity in resp.json():
            try:
                opportunity_processed = OpportunityRoot.model_validate(opportunity)
                opportunities.append(opportunity_processed.root)
            except UnsupportedOpportunityVersionException as e:
                warnings.warn(str(e))
        return opportunities

    async def submit_opportunity(self, opportunity: OpportunityParams) -> UUID:
        """
        Submits an opportunity to the server.

        Args:
            opportunity: An object representing the opportunity to submit.
        Returns:
            The ID of the submitted opportunity.
        """
        async with httpx.AsyncClient(**self.http_options) as client:
            resp = await client.post(
                urllib.parse.urlparse(self.server_url)
                ._replace(path="/v1/opportunities")
                .geturl(),
                json=opportunity.params.model_dump(),
            )
        resp.raise_for_status()
        return UUID(resp.json()["opportunity_id"])

    async def get_bids(self, from_time: datetime | None = None) -> list[BidResponse]:
        """
        Fetches bids for an api key from the server with pagination of 20 bids per page.

        Args:
            from_time: The datetime to fetch bids from. If None, fetches from the beginning of time.
        Returns:
            A list of bids.
        """
        async with httpx.AsyncClient(**self.http_options) as client:
            resp = await client.get(
                urllib.parse.urlparse(self.server_url)
                ._replace(path="/v1/bids")
                .geturl(),
                params=(
                    {"from_time": from_time.astimezone().isoformat()}
                    if from_time
                    else None
                ),
            )

        resp.raise_for_status()

        bids = []
        for bid in resp.json()["items"]:
            bid_processed = BidResponseRoot.model_validate(bid)
            bids.append(bid_processed.root)

        return bids

    @staticmethod
    def get_svm_submit_bid_instruction(
        searcher: Pubkey,
        router: Pubkey,
        permission_key: Pubkey,
        bid_amount: int,
        deadline: int,
        chain_id: str,
        fee_receiver_relayer: Pubkey,
        relayer_signer: Pubkey,
    ) -> Instruction:
        if chain_id not in SVM_CONFIGS:
            raise ValueError(f"Chain ID {chain_id} not supported")
        svm_config = SVM_CONFIGS[chain_id]
        config_router = LimoClient.get_express_relay_config_router_pda(
            SVM_EXPRESS_RELAY_PROGRAM_ID, router
        )
        express_relay_metadata = LimoClient.get_express_relay_metadata_pda(
            SVM_EXPRESS_RELAY_PROGRAM_ID
        )
        submit_bid_ix = submit_bid(
            {"data": SubmitBidArgs(deadline=deadline, bid_amount=bid_amount)},
            {
                "searcher": searcher,
                "relayer_signer": relayer_signer,
                "permission": permission_key,
                "router": router,
                "config_router": config_router,
                "express_relay_metadata": express_relay_metadata,
                "fee_receiver_relayer": fee_receiver_relayer,
                "sysvar_instructions": INSTRUCTIONS,
            },
            svm_config["express_relay_program"],
        )
        return submit_bid_ix


def compute_create2_address(
    searcher_address: Address,
    opportunity_adapter_factory_address: Address,
    opportunity_adapter_init_bytecode_hash: Bytes32,
) -> Address:
    """
    Computes the CREATE2 address for the opportunity adapter belonging to the searcher.

    Args:
        searcher_address: The address of the searcher's wallet.
        opportunity_adapter_factory_address: The address of the opportunity adapter factory.
        opportunity_adapter_init_bytecode_hash: The hash of the init code for the opportunity adapter.
    Returns:
        The computed CREATE2 address for the opportunity adapter.
    """
    pre = b"\xff"
    opportunity_adapter_factory = bytes.fromhex(
        opportunity_adapter_factory_address.replace("0x", "")
    )
    wallet = bytes.fromhex(searcher_address.replace("0x", ""))
    salt = bytes(12) + wallet
    init_code_hash = bytes.fromhex(
        opportunity_adapter_init_bytecode_hash.replace("0x", "")
    )
    result = web3.Web3.keccak(pre + opportunity_adapter_factory + salt + init_code_hash)
    return to_checksum_address(result[12:].hex())


def make_adapter_calldata(
    opportunity: OpportunityEvm,
    permitted: list[dict[str, Union[str, int]]],
    executor: Address,
    bid_params: OpportunityBidParams,
    signature: HexBytes,
):
    """
    Constructs the calldata for the opportunity adapter contract.

    Args:
        opportunity: An object representing the opportunity, of type Opportunity.
        permitted: A list of dictionaries representing the permitted tokens, in the format outputted by _get_permitted_tokens.
        executor: The address of the searcher's wallet.
        bid_params: An object representing the bid parameters, of type OpportunityBidParams.
        signature: The signature of the searcher's bid, as a HexBytes object.
    """
    function_selector = web3.Web3.solidity_keccak(
        ["string"], [f"executeOpportunity({EXECUTION_PARAMS_TYPESTRING},bytes)"]
    )[:4]
    function_args = encode(
        [EXECUTION_PARAMS_TYPESTRING, "bytes"],
        [
            (
                (
                    [(token["token"], token["amount"]) for token in permitted],
                    bid_params.nonce,
                    bid_params.deadline,
                ),
                (
                    [(token.token, token.amount) for token in opportunity.buy_tokens],
                    executor,
                    opportunity.target_contract,
                    bytes.fromhex(opportunity.target_calldata.replace("0x", "")),
                    opportunity.target_call_value,
                    bid_params.amount,
                ),
            ),
            signature,
        ],
    )
    calldata = f"0x{(function_selector + function_args).hex().replace('0x', '')}"
    return calldata


def get_opportunity_adapter_config(chain_id: str):
    opportunity_adapter_config = OPPORTUNITY_ADAPTER_CONFIGS.get(chain_id)
    if not opportunity_adapter_config:
        raise ExpressRelayClientException(
            f"Opportunity adapter config not found for chain id {chain_id}"
        )
    return opportunity_adapter_config


def get_signature(
    opportunity: OpportunityEvm,
    bid_params: OpportunityBidParams,
    private_key: str,
) -> SignedMessage:
    """
    Constructs a signature for a searcher's bid and opportunity.

    Args:
        opportunity: An object representing the opportunity, of type Opportunity.
        bid_params: An object representing the bid parameters, of type OpportunityBidParams.
        private_key: A 0x-prefixed hex string representing the searcher's private key.
    Returns:
        A SignedMessage object, representing the signature of the searcher's bid.
    """
    opportunity_adapter_config = get_opportunity_adapter_config(opportunity.chain_id)
    domain_data = {
        "name": "Permit2",
        "chainId": opportunity_adapter_config.chain_id,
        "verifyingContract": opportunity_adapter_config.permit2,
    }

    executor = Account.from_key(private_key).address
    message_types = {
        "PermitBatchWitnessTransferFrom": [
            {"name": "permitted", "type": "TokenPermissions[]"},
            {"name": "spender", "type": "address"},
            {"name": "nonce", "type": "uint256"},
            {"name": "deadline", "type": "uint256"},
            {"name": "witness", "type": "OpportunityWitness"},
        ],
        "OpportunityWitness": [
            {"name": "buyTokens", "type": "TokenAmount[]"},
            {"name": "executor", "type": "address"},
            {"name": "targetContract", "type": "address"},
            {"name": "targetCalldata", "type": "bytes"},
            {"name": "targetCallValue", "type": "uint256"},
            {"name": "bidAmount", "type": "uint256"},
        ],
        "TokenAmount": [
            {"name": "token", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "TokenPermissions": [
            {"name": "token", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
    }

    permitted = _get_permitted_tokens(
        opportunity.sell_tokens,
        bid_params.amount,
        opportunity.target_call_value,
        opportunity_adapter_config.weth,
    )

    # the data to be signed
    message_data = {
        "permitted": permitted,
        "spender": compute_create2_address(
            executor,
            opportunity_adapter_config.opportunity_adapter_factory,
            opportunity_adapter_config.opportunity_adapter_init_bytecode_hash,
        ),
        "nonce": bid_params.nonce,
        "deadline": bid_params.deadline,
        "witness": {
            "buyTokens": [
                {
                    "token": token.token,
                    "amount": int(token.amount),
                }
                for token in opportunity.buy_tokens
            ],
            "executor": executor,
            "targetContract": opportunity.target_contract,
            "targetCalldata": bytes.fromhex(
                opportunity.target_calldata.replace("0x", "")
            ),
            "targetCallValue": opportunity.target_call_value,
            "bidAmount": bid_params.amount,
        },
    }

    signed_typed_data = Account.sign_typed_data(
        private_key, domain_data, message_types, message_data
    )

    return signed_typed_data


def sign_opportunity_bid(
    opportunity: OpportunityEvm,
    bid_params: OpportunityBidParams,
    private_key: str,
) -> OpportunityBid:
    """
    Constructs a signature for a searcher's bid and returns the OpportunityBid object to be submitted to the server.

    Args:
        opportunity: An object representing the opportunity, of type Opportunity.
        bid_params: An object representing the bid parameters, of type OpportunityBidParams.
        private_key: A 0x-prefixed hex string representing the searcher's private key.
    Returns:
        A OpportunityBid object, representing the transaction to submit to the server. This object contains the searcher's signature.
    """
    executor = Account.from_key(private_key).address
    opportunity_bid = OpportunityBid(
        opportunity_id=opportunity.opportunity_id,
        permission_key=opportunity.permission_key,
        amount=bid_params.amount,
        deadline=bid_params.deadline,
        nonce=bid_params.nonce,
        executor=executor,
        signature=get_signature(opportunity, bid_params, private_key),
    )

    return opportunity_bid


def sign_bid(
    opportunity: OpportunityEvm, bid_params: OpportunityBidParams, private_key: str
) -> BidEvm:
    """
    Constructs a signature for a searcher's bid and returns the Bid object to be submitted to the server.

    Args:
        opportunity: An object representing the opportunity, of type Opportunity.
        bid_params: An object representing the bid parameters, of type OpportunityBidParams.
        private_key: A 0x-prefixed hex string representing the searcher's private key.
    Returns:
        A Bid object, representing the transaction to submit to the server. This object contains the searcher's signature.
    """
    opportunity_adapter_config = get_opportunity_adapter_config(opportunity.chain_id)
    permitted = _get_permitted_tokens(
        opportunity.sell_tokens,
        bid_params.amount,
        opportunity.target_call_value,
        opportunity_adapter_config.weth,
    )
    executor = Account.from_key(private_key).address

    signature = get_signature(opportunity, bid_params, private_key).signature
    calldata = make_adapter_calldata(
        opportunity, permitted, executor, bid_params, signature
    )

    return BidEvm(
        amount=bid_params.amount,
        target_calldata=calldata,
        chain_id=opportunity.chain_id,
        target_contract=opportunity_adapter_config.opportunity_adapter_factory,
        permission_key=opportunity.permission_key,
    )



import typing
from anchorpy.error import ProgramError


class FeeSplitLargerThanPrecision(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Fee split(s) larger than fee precision")

    code = 6000
    name = "FeeSplitLargerThanPrecision"
    msg = "Fee split(s) larger than fee precision"


class FeesHigherThanBid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Fees higher than bid")

    code = 6001
    name = "FeesHigherThanBid"
    msg = "Fees higher than bid"


class DeadlinePassed(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Deadline passed")

    code = 6002
    name = "DeadlinePassed"
    msg = "Deadline passed"


class InvalidCPISubmitBid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Invalid CPI into submit bid instruction")

    code = 6003
    name = "InvalidCPISubmitBid"
    msg = "Invalid CPI into submit bid instruction"


class MissingPermission(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Missing permission")

    code = 6004
    name = "MissingPermission"
    msg = "Missing permission"


class MultiplePermissions(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Multiple permissions")

    code = 6005
    name = "MultiplePermissions"
    msg = "Multiple permissions"


class InsufficientSearcherFunds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Insufficient Searcher Funds")

    code = 6006
    name = "InsufficientSearcherFunds"
    msg = "Insufficient Searcher Funds"


class InsufficientRent(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "Insufficient funds for rent")

    code = 6007
    name = "InsufficientRent"
    msg = "Insufficient funds for rent"


CustomError = typing.Union[
    FeeSplitLargerThanPrecision,
    FeesHigherThanBid,
    DeadlinePassed,
    InvalidCPISubmitBid,
    MissingPermission,
    MultiplePermissions,
    InsufficientSearcherFunds,
    InsufficientRent,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: FeeSplitLargerThanPrecision(),
    6001: FeesHigherThanBid(),
    6002: DeadlinePassed(),
    6003: InvalidCPISubmitBid(),
    6004: MissingPermission(),
    6005: MultiplePermissions(),
    6006: InsufficientSearcherFunds(),
    6007: InsufficientRent(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



import typing
from . import initialize_args
from .initialize_args import InitializeArgs, InitializeArgsJSON
from . import set_router_split_args
from .set_router_split_args import SetRouterSplitArgs, SetRouterSplitArgsJSON
from . import set_splits_args
from .set_splits_args import SetSplitsArgs, SetSplitsArgsJSON
from . import submit_bid_args
from .submit_bid_args import SubmitBidArgs, SubmitBidArgsJSON



import argparse
import asyncio
import logging
import typing

from eth_account.account import Account
from secrets import randbits

from express_relay.client import (
    ExpressRelayClient,
    sign_bid,
)
from express_relay.constants import OPPORTUNITY_ADAPTER_CONFIGS
from express_relay.models.evm import (
    BidEvm,
    Bytes32,
)
from express_relay.models import BidStatusUpdate, OpportunityBidParams, Opportunity
from express_relay.models.evm import OpportunityEvm, BidStatusEvm
from express_relay.models.base import BidStatusVariantsEvm

logger = logging.getLogger(__name__)

NAIVE_BID = int(1e16)
# Set deadline (naively) to max uint256
DEADLINE_MAX = 2**256 - 1


class SimpleSearcher:
    def __init__(
        self, server_url: str, private_key: Bytes32, api_key: str | None = None
    ):
        self.client = ExpressRelayClient(
            server_url,
            api_key,
            self.opportunity_callback,
            self.bid_status_callback,
        )
        self.private_key = private_key
        self.public_key = Account.from_key(private_key).address

    def assess_opportunity(
        self,
        opp: OpportunityEvm,
    ) -> BidEvm | None:
        """
        Assesses whether an opportunity is worth executing; if so, returns a Bid object.
        Otherwise, returns None.

        This function determines whether the given opportunity is worthwhile to execute.
        There are many ways to evaluate this, but the most common way is to check that the value of the tokens the searcher will receive from execution exceeds the value of the tokens spent.
        Individual searchers will have their own methods to determine market impact and the profitability of executing an opportunity. This function can use external prices to perform this evaluation.
        In this simple searcher, the function (naively) returns a Bid object with a default bid and deadline timestamp.
        Args:
            opp: An object representing a single opportunity.
        Returns:
            If the opportunity is deemed worthwhile, this function can return a Bid object, whose contents can be submitted to the auction server. If the opportunity is not deemed worthwhile, this function can return None.
        """

        # TODO: generate nonce more intelligently, to reduce gas costs
        bid_params = OpportunityBidParams(
            amount=NAIVE_BID, nonce=randbits(64), deadline=DEADLINE_MAX
        )

        bid = sign_bid(opp, bid_params, self.private_key)

        return bid

    async def opportunity_callback(self, opp: Opportunity):
        """
        Callback function to run when a new opportunity is found.

        Args:
            opp: An object representing a single opportunity.
        """
        bid = self.assess_opportunity(typing.cast(OpportunityEvm, opp))
        if bid:
            try:
                await self.client.submit_bid(bid)
                logger.info(
                    f"Submitted bid amount {bid.amount} for opportunity {str(opp.opportunity_id)}"
                )
            except Exception as e:
                logger.error(
                    f"Error submitting bid amount {bid.amount} for opportunity {str(opp.opportunity_id)}: {e}"
                )

    async def bid_status_callback(self, bid_status_update: BidStatusUpdate):
        """
        Callback function to run when a bid status is updated.

        Args:
            bid_status_update: An object representing an update to the status of a bid.
        """
        id = bid_status_update.id
        bid_status = typing.cast(BidStatusEvm, bid_status_update.bid_status)
        status = bid_status.type
        result = bid_status.result
        index = bid_status.index

        result_details = ""
        if status == BidStatusVariantsEvm.SUBMITTED or status == BidStatusVariantsEvm.WON:
            result_details = f", transaction {result}, index {index} of multicall"
        elif status == BidStatusVariantsEvm.LOST:
            if result:
                result_details = f", transaction {result}"
            if index:
                result_details += f", index {index} of multicall"
        logger.info(
            f"Bid status for bid {id}: {status.value.replace('_', ' ')}{result_details}"
        )


async def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument(
        "--private-key",
        type=str,
        required=True,
        help="Private key of the searcher for signing calldata as a hex string",
    )
    parser.add_argument(
        "--chain-ids",
        type=str,
        required=True,
        nargs="+",
        help="Chain ID(s) of the network(s) to monitor for opportunities",
    )
    parser.add_argument(
        "--server-url",
        type=str,
        required=True,
        help="Server endpoint to use for fetching opportunities and submitting bids",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        required=False,
        help="The API key of the searcher to authenticate with the server for fetching and submitting bids",
    )
    args = parser.parse_args()

    logger.setLevel(logging.INFO if args.verbose == 0 else logging.DEBUG)
    log_handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s:%(name)s:%(module)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    log_handler.setFormatter(formatter)
    logger.addHandler(log_handler)

    simple_searcher = SimpleSearcher(args.server_url, args.private_key, args.api_key)
    logger.info("Searcher address: %s", simple_searcher.public_key)
    for chain_id in args.chain_ids:
        if chain_id not in OPPORTUNITY_ADAPTER_CONFIGS:
            raise ValueError(
                f"Opportunity adapter config not found for chain {chain_id}"
            )
    await simple_searcher.client.subscribe_chains(args.chain_ids)

    task = await simple_searcher.client.get_ws_loop()
    await task


if __name__ == "__main__":
    asyncio.run(main())



from decimal import Decimal
from typing import Sequence, List, TypedDict, Tuple

from solana.constants import SYSTEM_PROGRAM_ID
from solana.rpc.async_api import AsyncClient
from solana.rpc.types import MemcmpOpts
from solders import system_program
from solders.instruction import Instruction, AccountMeta
from solders.pubkey import Pubkey
from solders.system_program import TransferParams
from solders.sysvar import RENT, INSTRUCTIONS
from spl.token._layouts import MINT_LAYOUT, ACCOUNT_LAYOUT as TOKEN_ACCOUNT_LAYOUT
from spl.token.constants import (
    WRAPPED_SOL_MINT,
    ASSOCIATED_TOKEN_PROGRAM_ID,
    TOKEN_PROGRAM_ID,
)

from express_relay.svm.generated.limo.accounts import Order
from express_relay.svm.generated.limo.instructions import (
    take_order,
    TakeOrderArgs,
)
from express_relay.svm.generated.limo.program_id import PROGRAM_ID

import spl.token.instructions as spl_token

ESCROW_VAULT_SEED = b"escrow_vault"
GLOBAL_AUTH_SEED = b"authority"
EXPRESS_RELAY_MEATADATA_SEED = b"metadata"
EXPRESS_RELAY_CONFIG_ROUTER_SEED = b"config_router"
EVENT_AUTH_SEED = b"__event_authority"
INTERMEDIARY_OUTPUT_TOKEN_ACCOUNT_SEED = b"intermediary"


class OrderStateAndAddress(TypedDict):
    state: Order
    address: Pubkey


class WSOLInstructions(TypedDict):
    create_ixs: List[Instruction]
    fill_ixs: List[Instruction]
    close_ixs: List[Instruction]
    ata: Pubkey


class LimoClient:
    def __init__(self, connection: AsyncClient):
        self._connection = connection

    async def get_all_orders_state_and_address_with_filters(
        self, filters: List[MemcmpOpts], global_config: Pubkey
    ) -> List[OrderStateAndAddress]:
        filters.append(MemcmpOpts(offset=8, bytes=str(global_config)))
        programs = await self._connection.get_program_accounts(
            PROGRAM_ID,
            commitment=None,
            encoding="base64",
            data_slice=None,
            filters=filters,
        )

        return [
            {"state": Order.decode(value.account.data), "address": value.pubkey}
            for value in programs.value
        ]

    async def get_mint_decimals(self, mint: Pubkey) -> int:
        mint_account = await self._connection.get_account_info(mint)
        if mint_account.value is None:
            raise ValueError("Mint account not found")
        bytes_data = mint_account.value.data
        if len(bytes_data) != MINT_LAYOUT.sizeof():
            raise ValueError("Invalid mint size")

        decoded_data = MINT_LAYOUT.parse(bytes_data)
        decimals = decoded_data.decimals
        return decimals

    async def account_exists(self, address: Pubkey) -> bool:
        account_info = await self._connection.get_account_info(address)
        return account_info.value is not None

    def create_associated_token_account_idempotent(
        self, payer: Pubkey, owner: Pubkey, mint: Pubkey, token_program_id: Pubkey
    ) -> Instruction:
        """Creates a transaction instruction to create an associated token account.

        Returns:
            The instruction to create the associated token account.
        """
        associated_token_address = self.get_ata(owner, mint, token_program_id)
        return Instruction(
            accounts=[
                AccountMeta(pubkey=payer, is_signer=True, is_writable=True),
                AccountMeta(
                    pubkey=associated_token_address, is_signer=False, is_writable=True
                ),
                AccountMeta(pubkey=owner, is_signer=False, is_writable=False),
                AccountMeta(pubkey=mint, is_signer=False, is_writable=False),
                AccountMeta(
                    pubkey=SYSTEM_PROGRAM_ID, is_signer=False, is_writable=False
                ),
                AccountMeta(
                    pubkey=token_program_id, is_signer=False, is_writable=False
                ),
                AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
            ],
            program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
            data=bytes([1]),  # idempotent version of the instruction
        )

    async def get_ata_and_create_ixn_if_required(
        self,
        owner: Pubkey,
        token_mint_address: Pubkey,
        token_program_id: Pubkey,
        payer: Pubkey,
    ) -> Tuple[Pubkey, Sequence[Instruction]]:
        ata = self.get_ata(owner, token_mint_address, token_program_id)
        if not await self.account_exists(ata):
            ix = self.create_associated_token_account_idempotent(
                payer, owner, token_mint_address, token_program_id
            )
            return ata, [ix]
        return ata, []

    async def get_init_if_needed_wsol_create_and_close_ixs(
        self, owner: Pubkey, payer: Pubkey, amount_to_deposit_lamports: int
    ) -> WSOLInstructions:
        """
        Returns necessary instructions to create, fill and close a wrapped SOL account.
        If the account already exists:
            it makes sure the balance is at least `amount_to_deposit_lamports` with no create or close instructions.
        If the account does not exist:
            it creates the account, fills it with `amount_to_deposit_lamports` lamports and close it in the end.
        Args:
            owner: Who owns the WSOL token account
            payer: Who pays for the instructions
            amount_to_deposit_lamports: Minimum amount of lamports required in the account
        """
        ata = self.get_ata(owner, WRAPPED_SOL_MINT, TOKEN_PROGRAM_ID)
        ata_info = await self._connection.get_account_info(ata)

        create_ixs = []
        close_ixs = []
        if ata_info.value is None or len(ata_info.value.data) == 0:
            create_ixs = [
                self.create_associated_token_account_idempotent(
                    payer, owner, WRAPPED_SOL_MINT, TOKEN_PROGRAM_ID
                )
            ]
            close_ixs = [
                spl_token.close_account(
                    spl_token.CloseAccountParams(
                        program_id=TOKEN_PROGRAM_ID,
                        account=ata,
                        dest=owner,
                        owner=owner,
                    )
                )
            ]

        fill_ixs = []
        current_balance = (
            TOKEN_ACCOUNT_LAYOUT.parse(ata_info.value.data).amount
            if ata_info.value and len(ata_info.value.data) > 0
            else 0
        )
        if current_balance < amount_to_deposit_lamports:
            fill_ixs = [
                system_program.transfer(
                    TransferParams(
                        from_pubkey=owner,
                        to_pubkey=ata,
                        lamports=amount_to_deposit_lamports - current_balance,
                    )
                ),
                spl_token.sync_native(
                    spl_token.SyncNativeParams(TOKEN_PROGRAM_ID, ata)
                ),
            ]

        return WSOLInstructions(
            create_ixs=create_ixs, fill_ixs=fill_ixs, close_ixs=close_ixs, ata=ata
        )

    async def take_order_ix(
        self,
        taker: Pubkey,
        order: OrderStateAndAddress,
        input_amount: int,
        output_amount: int,
        express_relay_program_id: Pubkey,
    ) -> List[Instruction]:
        """
        Returns the instructions to fulfill an order as a taker.
        Args:
            taker: The taker's public key
            order: The order to fulfill
            input_amount_decimals: The amount of input tokens to take. Will be multiplied by 10 ** input_mint_decimals in this method.
            output_amount_decimals: The amount of output tokens to provide. Will be multiplied by 10 ** output_mint_decimals in this method.
            input_mint_decimals: input mint decimals (can be fetched via get_mint_decimals)
            output_mint_decimals: output mint decimals (can be fetched via get_mint_decimals)
            express_relay_program_id: Express relay program id

        Returns:
            A list of instructions to include in the transaction to fulfill the order. The submit_bid instruction for
            express relay program is not included and should be added separately.

        """
        ixs: List[Instruction] = []
        close_wsol_ixns: List[Instruction] = []
        taker_input_ata: Pubkey
        if order["state"].input_mint == WRAPPED_SOL_MINT:
            instructions = await self.get_init_if_needed_wsol_create_and_close_ixs(
                owner=taker, payer=taker, amount_to_deposit_lamports=0
            )
            ixs.extend(instructions["create_ixs"])
            close_wsol_ixns.extend(instructions["close_ixs"])
            taker_input_ata = instructions["ata"]
        else:
            (
                taker_input_ata,
                create_taker_input_ata_ixs,
            ) = await self.get_ata_and_create_ixn_if_required(
                owner=taker,
                token_mint_address=order["state"].input_mint,
                token_program_id=order["state"].input_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_taker_input_ata_ixs)

        taker_output_ata: Pubkey
        maker_output_ata: Pubkey | None = None
        intermediary_output_token_account: Pubkey | None = None
        if order["state"].output_mint == WRAPPED_SOL_MINT:
            instructions = await self.get_init_if_needed_wsol_create_and_close_ixs(
                owner=taker, payer=taker, amount_to_deposit_lamports=output_amount
            )
            ixs.extend(instructions["create_ixs"])
            ixs.extend(instructions["fill_ixs"])
            close_wsol_ixns.extend(instructions["close_ixs"])
            taker_output_ata = instructions["ata"]

            intermediary_output_token_account = self.get_intermediary_token_account_pda(PROGRAM_ID, order["address"])
        else:
            (
                taker_output_ata,
                create_taker_output_ata_ixs,
            ) = await self.get_ata_and_create_ixn_if_required(
                owner=taker,
                token_mint_address=order["state"].output_mint,
                token_program_id=order["state"].output_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_taker_output_ata_ixs)

            (
                maker_output_ata,
                create_maker_output_ata_ixs,
            ) = await self.get_ata_and_create_ixn_if_required(
                owner=order["state"].maker,
                token_mint_address=order["state"].output_mint,
                token_program_id=order["state"].output_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_maker_output_ata_ixs)

        pda_authority = self.get_pda_authority(PROGRAM_ID, order["state"].global_config)
        ixs.append(
            take_order(
                TakeOrderArgs(
                    input_amount=input_amount,
                    min_output_amount=output_amount,
                    tip_amount_permissionless_taking=0,
                ),
                {
                    "taker": taker,
                    "maker": order["state"].maker,
                    "global_config": order["state"].global_config,
                    "pda_authority": pda_authority,
                    "order": order["address"],
                    "input_mint": order["state"].input_mint,
                    "output_mint": order["state"].output_mint,
                    "input_vault": self.get_token_vault_pda(
                        PROGRAM_ID,
                        order["state"].global_config,
                        order["state"].input_mint,
                    ),
                    "taker_input_ata": taker_input_ata,
                    "taker_output_ata": taker_output_ata,
                    "intermediary_output_token_account": intermediary_output_token_account,
                    "maker_output_ata": maker_output_ata,
                    "express_relay": express_relay_program_id,
                    "express_relay_metadata": self.get_express_relay_metadata_pda(
                        express_relay_program_id
                    ),
                    "sysvar_instructions": INSTRUCTIONS,
                    "permission": order["address"],
                    "config_router": self.get_express_relay_config_router_pda(
                        express_relay_program_id, pda_authority
                    ),
                    "input_token_program": order["state"].input_mint_program_id,
                    "output_token_program": order["state"].output_mint_program_id,
                    "event_authority": self.get_event_authority(PROGRAM_ID),
                    "program": PROGRAM_ID,
                },
            )
        )

        ixs.extend(close_wsol_ixns)
        return ixs

    @staticmethod
    def get_program_id() -> Pubkey:
        return PROGRAM_ID

    @staticmethod
    def get_token_vault_pda(
        program_id: Pubkey, global_config: Pubkey, input_mint: Pubkey
    ) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[ESCROW_VAULT_SEED, bytes(global_config), bytes(input_mint)],
            program_id=program_id,
        )[0]

    @staticmethod
    def get_express_relay_metadata_pda(program_id: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EXPRESS_RELAY_MEATADATA_SEED], program_id=program_id
        )[0]

    @staticmethod
    def get_express_relay_config_router_pda(
        program_id: Pubkey, router: Pubkey
    ) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EXPRESS_RELAY_CONFIG_ROUTER_SEED, bytes(router)],
            program_id=program_id,
        )[0]

    @staticmethod
    def get_pda_authority(program_id: Pubkey, global_config: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[GLOBAL_AUTH_SEED, bytes(global_config)], program_id=program_id
        )[0]

    @staticmethod
    def get_ata(
        owner: Pubkey, token_mint_address: Pubkey, token_program_id: Pubkey
    ) -> Pubkey:
        ata, _ = Pubkey.find_program_address(
            seeds=[bytes(owner), bytes(token_program_id), bytes(token_mint_address)],
            program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
        )
        return ata

    @staticmethod
    def get_event_authority(program_id: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EVENT_AUTH_SEED], program_id=program_id
        )[0]

    @staticmethod
    def get_intermediary_token_account_pda(program_id: Pubkey, order_address: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[INTERMEDIARY_OUTPUT_TOKEN_ACCOUNT_SEED, bytes(order_address)], program_id=program_id
        )[0]



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class SetRouterSplitArgs(typing.TypedDict):
    data: types.set_router_split_args.SetRouterSplitArgs


layout = borsh.CStruct("data" / types.set_router_split_args.SetRouterSplitArgs.layout)


class SetRouterSplitAccounts(typing.TypedDict):
    admin: Pubkey
    config_router: Pubkey
    express_relay_metadata: Pubkey
    router: Pubkey


def set_router_split(
    args: SetRouterSplitArgs,
    accounts: SetRouterSplitAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["config_router"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(pubkey=accounts["router"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x10\x96j\r\x1b\xbfh\x08"
    encoded_args = layout.build(
        {
            "data": args["data"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class InitializeArgs(typing.TypedDict):
    data: types.initialize_args.InitializeArgs


layout = borsh.CStruct("data" / types.initialize_args.InitializeArgs.layout)


class InitializeAccounts(typing.TypedDict):
    payer: Pubkey
    express_relay_metadata: Pubkey
    admin: Pubkey
    relayer_signer: Pubkey
    fee_receiver_relayer: Pubkey


def initialize(
    args: InitializeArgs,
    accounts: InitializeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["admin"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["relayer_signer"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["fee_receiver_relayer"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xaf\xafm\x1f\r\x98\x9b\xed"
    encoded_args = layout.build(
        {
            "data": args["data"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class CloseOrderAndClaimTipAccounts(typing.TypedDict):
    maker: Pubkey
    order: Pubkey
    global_config: Pubkey
    pda_authority: Pubkey
    input_mint: Pubkey
    output_mint: Pubkey
    maker_input_ata: Pubkey
    input_vault: Pubkey
    input_token_program: Pubkey
    event_authority: Pubkey
    program: Pubkey


def close_order_and_claim_tip(
    accounts: CloseOrderAndClaimTipAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["maker"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["order"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["pda_authority"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["input_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["output_mint"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["maker_input_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["input_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["input_token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["event_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["program"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf4\x1b\x0c\xe2-\xf7\xe6+"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class WithdrawFeesAccounts(typing.TypedDict):
    admin: Pubkey
    fee_receiver_admin: Pubkey
    express_relay_metadata: Pubkey


def withdraw_fees(
    accounts: WithdrawFeesAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["fee_receiver_admin"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xc6\xd4\xabm\x90\xd7\xaeY"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class SetSplitsArgs(typing.TypedDict):
    data: types.set_splits_args.SetSplitsArgs


layout = borsh.CStruct("data" / types.set_splits_args.SetSplitsArgs.layout)


class SetSplitsAccounts(typing.TypedDict):
    admin: Pubkey
    express_relay_metadata: Pubkey


def set_splits(
    args: SetSplitsArgs,
    accounts: SetSplitsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xaf\x02V1\xe1\xca\xe8\xbd"
    encoded_args = layout.build(
        {
            "data": args["data"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeVaultAccounts(typing.TypedDict):
    admin_authority: Pubkey
    global_config: Pubkey
    pda_authority: Pubkey
    mint: Pubkey
    vault: Pubkey


def initialize_vault(
    accounts: InitializeVaultAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["admin_authority"], is_signer=True, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["pda_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"0\xbf\xa3,G\x81?\xa4"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
